"""Constants for gate devices"""
from enum import IntEnum

from typing_extensions import Final


class GogoGate2ApiErrorCode(IntEnum):
    """Common API error codes."""

    CREDENTIALS_NOT_SET = 2
    CREDENTIALS_INCORRECT = 1
    INVALID_OPTION = 9
    CORRUPTED_DATA = 11

    INVALID_API_CODE = 18
    DOOR_NOT_SET = 8
    INVALID_DOOR = 5

    # These are not thrown by the api.
    INVALID_TOKEN = 998
    TOKEN_NOT_SET = 999


class ISmartGateApiErrorCode(IntEnum):
    """Common API error codes."""

    CREDENTIALS_NOT_SET = 22
    CREDENTIALS_INCORRECT = 11
    INVALID_OPTION = 9
    TOKEN_NOT_SET = 21
    INVALID_API_CODE = 10
    DOOR_NOT_SET = 8

    # These are not thrown by the api.
    INVALID_DOOR = 997
    INVALID_TOKEN = 998
    CORRUPTED_DATA = 999


NONE_INT: Final = -100000



"""Common code for gate APIs."""
import dataclasses
import json
from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from typing import Any, Callable, Optional, Tuple, Type, TypeVar, Union, cast
from xml.etree.ElementTree import Element  # nosec

from typing_extensions import Final

from .const import NONE_INT

GenericType = TypeVar("GenericType")


class InvalidDoorException(Exception):
    """Exception when door is invalid."""

    def __init__(self, door_id: int) -> None:
        super().__init__(f"Door with id {door_id} not found.")
        self.door_id: Final = door_id


class TagException(Exception):
    """General exception for tags."""

    def __init__(self, tag: str, message: str) -> None:
        super().__init__(message)
        self.tag: Final = tag


class TagNotFoundException(TagException):
    """Thrown when encountering an unexpected type."""

    def __init__(self, tag: str) -> None:
        """Initialize."""
        super().__init__(tag, f"Did not find element tag '{tag}'.")


class TextEmptyException(TagException):
    """Thrown when encountering an unexpected type."""

    def __init__(self, tag: str) -> None:
        """Initialize."""
        super().__init__(tag, f"Text was empty for tag '{tag}'.")


class UnexpectedTypeException(Exception):
    """Thrown when encountering an unexpected type."""

    def __init__(self, value: Any, expected: Type[GenericType]):
        """Initialize."""
        super().__init__(
            f'Expected of "{value}" to be "{expected}" but was "{type(value)}."'
        )
        self.value: Final = value
        self.expected: Final = expected


def enforce_type(value: Any, expected: Type[GenericType]) -> GenericType:
    """Enforce a data type."""
    if not isinstance(value, expected):
        raise UnexpectedTypeException(value, expected)

    return value


def value_or_none(
    value: Any, convert_fn: Callable[[Any], GenericType]
) -> Union[GenericType, None]:
    """Convert a value given a specific conversion function."""
    if value is None:
        return None

    try:
        return convert_fn(value)
    except Exception:  # pylint: disable=broad-except
        return None


def enum_or_raise(value: Optional[Union[str, int]], enum: Type[Enum]) -> Enum:
    """Return Enum or raise exception."""
    if value is None:
        raise UnexpectedTypeException(value, enum)

    return enum(value)


def str_or_raise(value: Any) -> str:
    """Return string or raise exception."""
    return enforce_type(str_or_none(value), str)


def str_or_none(value: Any) -> Optional[str]:
    """Return str or None."""
    return value_or_none(value, str)


def int_or_raise(value: Any) -> int:
    """Return int or raise exception."""
    return enforce_type(int_or_none(value), int)


def int_or_none(value: Any) -> Optional[int]:
    """Return int or None."""
    return value_or_none(value, int)


def float_or_none(value: Any) -> Optional[float]:
    """Return float or None."""
    return value_or_none(value, float)


class EnhancedJSONEncoder(json.JSONEncoder):
    """JSON encoder."""

    # pylint: disable=method-hidden
    def default(self, o) -> Any:  # type: ignore
        """Encode the object."""
        if dataclasses.is_dataclass(o):
            return dataclasses.asdict(o)
        if isinstance(o, Enum):
            return o.value
        return super().default(o)


class ApiError(Exception):
    """Generic API error."""

    def __init__(self, code: int, message: str) -> None:
        super().__init__(f"Code: {code} - {message}")
        self._code: Final = code
        self._message: Final = message

    @property
    def code(self) -> int:
        """Get the error code."""
        return self._code

    @property
    def message(self) -> str:
        """Get the error message."""
        return self._message


class CredentialsNotSetException(ApiError):
    """Credentials not set exception."""


class CredentialsIncorrectException(ApiError):
    """Credentials incorrect exception."""


class InvalidOptionException(ApiError):
    """Invalid option exception."""


class InvalidApiCodeException(ApiError):
    """Invalid api code exception."""


class DoorNotSetException(ApiError):
    """Door not set exception."""


ExceptionGenerator: Final = Callable[[int, str], ApiError]


class DoorStatus(Enum):
    """Door status."""

    CLOSED = "closed"
    OPENED = "opened"
    UNDEFINED = "undefined"


class TransitionDoorStatus(Enum):
    """Presumed door status."""

    OPENING = "opening"
    CLOSING = "closing"


AllDoorStatus: Final = Union[DoorStatus, TransitionDoorStatus]


CLOSE_DOOR_STATUSES: Final = frozenset(
    (DoorStatus.CLOSED, TransitionDoorStatus.CLOSING)
)
OPEN_DOOR_STATUSES: Final = frozenset((DoorStatus.OPENED, TransitionDoorStatus.OPENING))


class DoorMode(Enum):
    """Door mode."""

    GARAGE = "garage"
    PULSE = "pulse"
    ON_OFF = "onoff"


class RequestOption(Enum):
    """Request option."""

    INFO = "info"
    ACTIVATE = "activate"


@dataclass(frozen=True)
# pylint: disable=too-many-instance-attributes
class AbstractDoor:
    """Door object."""

    door_id: int
    permission: bool
    name: Optional[str]
    mode: DoorMode
    gate: bool
    status: DoorStatus
    sensor: bool
    camera: bool
    events: Optional[int]
    sensorid: Optional[str]
    temperature: Optional[float]
    voltage: Optional[int]


@dataclass(frozen=True)
class GogoGate2Door(AbstractDoor):
    """Door object."""


@dataclass(frozen=True)
class ISmartGateDoor(AbstractDoor):
    """Door object."""

    enabled: bool
    apicode: str
    customimage: bool


@dataclass(frozen=True)
class Outputs:
    """Outputs object."""

    output1: bool
    output2: bool
    output3: bool


@dataclass(frozen=True)
class Network:
    """Network object."""

    ip: str  # pylint: disable=invalid-name


@dataclass(frozen=True)
class Wifi:
    """Wifi object."""

    SSID: Optional[str]  # pylint: disable=invalid-name
    linkquality: Optional[str]
    signal: Optional[str]


@dataclass(frozen=True)
class AbstractActivateResponse:
    """Response from activate calls."""

    result: bool


@dataclass(frozen=True)
class GogoGate2ActivateResponse(AbstractActivateResponse):
    """GogoGate2 response from activate calls."""


@dataclass(frozen=True)
class ISmartGateActivateResponse(AbstractActivateResponse):
    """iSmartGate response from activate calls."""


@dataclass(frozen=True)
# pylint: disable=too-many-instance-attributes
class AbstractInfoResponse:
    """Response from info calls."""

    user: str
    model: str
    apiversion: str
    remoteaccessenabled: bool
    remoteaccess: str
    firmwareversion: str

    door1: AbstractDoor
    door2: AbstractDoor
    door3: AbstractDoor

    network: Network
    wifi: Wifi


@dataclass(frozen=True)
#  pylint: disable=too-many-instance-attributes
class GogoGate2InfoResponse(AbstractInfoResponse):
    """Response from gogogate2 api calls."""

    gogogatename: str
    apicode: str
    door1: GogoGate2Door
    door2: GogoGate2Door
    door3: GogoGate2Door
    outputs: Outputs


@dataclass(frozen=True)
class ISmartGateInfoResponse(AbstractInfoResponse):
    """Response from iSmartGate info calls."""

    pin: int
    lang: str
    ismartgatename: str
    newfirmware: bool

    door1: ISmartGateDoor
    door2: ISmartGateDoor
    door3: ISmartGateDoor


@dataclass(frozen=True)
class CachedTransitionDoorStatus:
    """Cached persistent door status."""

    door_id: int
    activated: datetime
    transition_status: TransitionDoorStatus
    target_status: DoorStatus


def element_or_none(element: Optional[Element], tag: str) -> Optional[Element]:
    """Get element from xml element."""
    return None if element is None else element.find(tag)


def element_or_raise(element: Optional[Element], tag: str) -> Element:
    """Get element from xml element."""
    found_element: Final = element_or_none(element, tag)
    if found_element is None:
        raise TagNotFoundException(tag)

    return found_element


def element_text_or_none(element: Optional[Element], tag: str) -> Optional[str]:
    """Get element text from xml element."""
    found_element: Final = element_or_none(element, tag)
    return (
        None
        if found_element is None
        else None
        if found_element.text is None
        else found_element.text.strip()
    )


def element_text_or_raise(element: Optional[Element], tag: str) -> str:
    """Get element text from xml element."""
    found_element: Final = element_or_raise(element, tag)
    if found_element.text is None:
        raise TextEmptyException(tag)

    return found_element.text.strip()


def element_int_or_raise(element: Optional[Element], tag: str) -> int:
    """Get element int from xml element."""
    found_element: Final = element_or_raise(element, tag)
    if found_element.text is None:
        raise TextEmptyException(tag)

    return int_or_raise(found_element.text.strip())


def element_to_api_error(element: Element) -> ApiError:
    """Get api error from xml element."""
    return ApiError(
        code=int_or_raise(element_text_or_raise(element, "errorcode")),
        message=element_text_or_raise(element, "errormsg"),
    )


def wifi_or_raise(element: Element) -> Wifi:
    """Get wifi from xml element."""
    return Wifi(
        SSID=element_text_or_none(element, "SSID"),
        linkquality=element_text_or_none(element, "linkquality"),
        signal=element_text_or_none(element, "signal"),
    )


def network_or_raise(element: Element) -> Network:
    """Get network from xml element."""
    return Network(
        ip=element_text_or_raise(element, "ip"),
    )


def outputs_or_raise(element: Element) -> Outputs:
    """Get outputs from xml element."""
    return Outputs(
        output1=element_text_or_raise(element, "output1").lower() == "on",
        output2=element_text_or_raise(element, "output2").lower() == "on",
        output3=element_text_or_raise(element, "output3").lower() == "on",
    )


def gogogate2_door_or_raise(door_id: int, element: Element) -> GogoGate2Door:
    """Get door from xml element."""
    temp: Final = float_or_none(element_text_or_none(element, "temperature"))
    voltage: Final = int_or_none(element_text_or_none(element, "voltage"))
    return GogoGate2Door(
        door_id=door_id,
        permission=element_text_or_raise(element, "permission").lower() == "yes",
        name=element_text_or_none(element, "name"),
        gate=False,
        mode=cast(
            DoorMode, enum_or_raise(element_text_or_raise(element, "mode"), DoorMode)
        ),
        status=cast(
            DoorStatus,
            enum_or_raise(element_text_or_raise(element, "status"), DoorStatus),
        ),
        sensor=element_text_or_raise(element, "sensor").lower() == "yes",
        sensorid=element_text_or_none(element, "sensorid"),
        camera=element_text_or_raise(element, "camera").lower() == "yes",
        events=int_or_none(element_text_or_none(element, "events")),
        temperature=None if temp is None else None if temp <= NONE_INT else temp,
        voltage=None if voltage is None else None if voltage <= NONE_INT else voltage,
    )


def ismartgate_door_or_raise(door_id: int, element: Element) -> ISmartGateDoor:
    """Get door from xml element."""
    temp: Final = float_or_none(element_text_or_none(element, "temperature"))
    voltage: Final = int_or_none(element_text_or_none(element, "voltage"))
    return ISmartGateDoor(
        door_id=door_id,
        enabled=element_text_or_raise(element, "enabled").lower() == "yes",
        apicode=element_text_or_raise(element, "apicode"),
        customimage=element_text_or_raise(element, "customimage").lower() == "yes",
        permission=element_text_or_raise(element, "permission").lower() == "yes",
        name=element_text_or_none(element, "name"),
        gate=element_text_or_raise(element, "gate").lower() == "yes",
        mode=cast(
            DoorMode, enum_or_raise(element_text_or_raise(element, "mode"), DoorMode)
        ),
        status=cast(
            DoorStatus,
            enum_or_raise(element_text_or_raise(element, "status"), DoorStatus),
        ),
        sensor=element_text_or_raise(element, "sensor").lower() == "yes",
        sensorid=element_text_or_none(element, "sensorid"),
        camera=element_text_or_raise(element, "camera").lower() == "yes",
        events=int_or_none(element_text_or_none(element, "events")),
        temperature=None if temp is None else None if temp <= NONE_INT else temp,
        voltage=None if voltage is None else None if voltage <= NONE_INT else voltage,
    )


def element_to_gogogate2_info_response(element: Element) -> GogoGate2InfoResponse:
    """Get response from xml element."""
    return GogoGate2InfoResponse(
        user=element_text_or_raise(element, "user"),
        gogogatename=element_text_or_raise(element, "gogogatename"),
        model=element_text_or_raise(element, "model"),
        apiversion=element_text_or_raise(element, "apiversion"),
        remoteaccessenabled=element_text_or_raise(element, "remoteaccessenabled")
        == "1",
        remoteaccess=element_text_or_raise(element, "remoteaccess"),
        firmwareversion=element_text_or_raise(element, "firmwareversion"),
        apicode=element_text_or_raise(element, "apicode"),
        door1=gogogate2_door_or_raise(1, element_or_raise(element, "door1")),
        door2=gogogate2_door_or_raise(2, element_or_raise(element, "door2")),
        door3=gogogate2_door_or_raise(3, element_or_raise(element, "door3")),
        outputs=outputs_or_raise(element_or_raise(element, "outputs")),
        network=network_or_raise(element_or_raise(element, "network")),
        wifi=wifi_or_raise(element_or_raise(element, "wifi")),
    )


def element_to_ismartgate_info_response(element: Element) -> ISmartGateInfoResponse:
    """Get response from xml element."""
    return ISmartGateInfoResponse(
        user=element_text_or_raise(element, "user"),
        pin=element_int_or_raise(element, "pin"),
        lang=element_text_or_raise(element, "lang"),
        ismartgatename=element_text_or_raise(element, "ismartgatename"),
        model=element_text_or_raise(element, "model"),
        apiversion=element_text_or_raise(element, "apiversion"),
        remoteaccessenabled=element_text_or_raise(
            element, "remoteaccessenabled"
        ).lower()
        == "yes",
        remoteaccess=element_text_or_raise(element, "remoteaccess"),
        firmwareversion=element_text_or_raise(element, "firmwareversion"),
        newfirmware=element_text_or_raise(element, "newfirmware").lower() == "yes",
        door1=ismartgate_door_or_raise(1, element_or_raise(element, "door1")),
        door2=ismartgate_door_or_raise(2, element_or_raise(element, "door2")),
        door3=ismartgate_door_or_raise(3, element_or_raise(element, "door3")),
        network=network_or_raise(element_or_raise(element, "network")),
        wifi=wifi_or_raise(element_or_raise(element, "wifi")),
    )


def element_to_gogogate2_activate_response(
    element: Element,
) -> GogoGate2ActivateResponse:
    """Get response from xml element."""
    return GogoGate2ActivateResponse(
        result=element_text_or_raise(element, "result").lower() == "ok"
    )


def element_to_ismartgate_activate_response(
    element: Element,
) -> ISmartGateActivateResponse:
    """Get response from xml element."""
    return ISmartGateActivateResponse(
        result=element_text_or_raise(element, "result").lower() == "ok"
    )


def get_door_by_id(
    door_id: int, response: AbstractInfoResponse
) -> Optional[AbstractDoor]:
    """Get a door from a gogogate2 response."""
    return next(
        iter([door for door in get_doors(response) if door.door_id == door_id]), None
    )


def get_doors(response: AbstractInfoResponse) -> Tuple[AbstractDoor, ...]:
    """Get a tuple of doors from a response."""
    return (response.door1, response.door2, response.door3)


def get_configured_door_by_id(
    door_id: int, response: AbstractInfoResponse
) -> Optional[AbstractDoor]:
    """Get a door from a gogogate2 response."""
    return next(
        iter(
            [door for door in get_configured_doors(response) if door.door_id == door_id]
        ),
        None,
    )


def get_configured_doors(response: AbstractInfoResponse) -> Tuple[AbstractDoor, ...]:
    """Get a tuple of configured doors from a response."""
    return tuple(door for door in get_doors(response) if door.name)



"""CLI for gate devices."""
import asyncio
import json
import pprint
from enum import Enum, unique
from functools import wraps
from getpass import getpass
from typing import Any, Callable, Union, cast

import click
from typing_extensions import Final

from . import AbstractGateApi, GogoGate2Api, ISmartGateApi
from .common import EnhancedJSONEncoder

API: Final = "api"
DEVICE_TYPE: Final = "device_type"
PRETTY_PRINT: Final = pprint.PrettyPrinter(indent=4)


def coro(func: Callable) -> Any:
    """Wrap a coroutine in a async runner."""

    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        return asyncio.run(func(*args, **kwargs))

    return wrapper


def _echo_response(obj: Any) -> None:
    click.echo(json.dumps(obj, indent=2, cls=EnhancedJSONEncoder))


@unique
class DeviceType(Enum):
    """Device types."""

    GOGOGATE2 = "gogogate2"
    ISMARTGATE = "ismartgate"


@unique
class Option(Enum):
    """CLI options."""

    HOST = "--host"
    USERNAME = "--username"
    PASSWORD = "--password"  # nosec
    DEVICE_TYPE = "--device-type"


@unique
class Command(Enum):
    """Device actions."""

    OPEN = "open"
    CLOSE = "close"
    INFO = "info"


@unique
class DoorArgument(Enum):
    """Door arguments."""

    DOOR_ID = "door_id"


def default_password() -> str:
    """Get the password from user input."""
    return getpass("Password: ").strip()


def get_context_api(ctx: click.core.Context) -> AbstractGateApi:
    """Get gate API from context."""

    return cast(AbstractGateApi, ctx.obj[API])


@click.group()
@click.option(Option.HOST.value, required=True)
@click.option(Option.USERNAME.value, required=True)
@click.option(
    Option.PASSWORD.value,
    default=default_password,
    help="Omit for interactive prompt. Use '-' to read from stdin.",
)
@click.option(
    Option.DEVICE_TYPE.value,
    required=True,
    type=click.Choice([item.value for item in DeviceType], case_sensitive=False),
    hidden=True,
)
@click.version_option()
@click.pass_context
def cli(
    ctx: click.core.Context,
    host: str,
    username: str,
    password: str,
    device_type: str,
) -> None:
    """Interact with the device API."""
    api_generator: Callable[[str, str, str], Union[GogoGate2Api, ISmartGateApi]]

    if device_type == DeviceType.GOGOGATE2.value:
        api_generator = GogoGate2Api
    else:
        api_generator = ISmartGateApi

    if password == "-":  # nosec
        password = default_password()

    ctx.obj = {API: api_generator(host, username, password)}


@cli.command(name=Command.INFO.value)
@click.pass_context
@coro
async def info(ctx: click.core.Context) -> None:
    """Get info from device."""
    _echo_response(await get_context_api(ctx).async_info())


@cli.command(name=Command.OPEN.value)
@click.argument(DoorArgument.DOOR_ID.value, type=int, required=True)
@click.pass_context
@coro
async def open_door(ctx: click.core.Context, door_id: int) -> None:
    """Open the door."""
    _echo_response(await get_context_api(ctx).async_open_door(door_id))


@cli.command(name=Command.CLOSE.value)
@click.argument(DoorArgument.DOOR_ID.value, type=int, required=True)
@click.pass_context
@coro
async def close_door(ctx: click.core.Context, door_id: int) -> None:
    """Close the door."""
    _echo_response(await get_context_api(ctx).async_close_door(door_id))


def cli_with_defaults(device_type: DeviceType) -> None:
    """Run command with default value."""
    cli(  # pylint: disable=no-value-for-parameter,unexpected-keyword-arg
        default_map={DEVICE_TYPE: device_type.value}
    )


def gogogate2_cli() -> None:
    """GogoGate2 entrypoint."""

    cli_with_defaults(DeviceType.GOGOGATE2)


def ismartgate_cli() -> None:
    """iSmartGate entrypoint."""
    cli_with_defaults(DeviceType.ISMARTGATE)


def maybe_init() -> None:
    """Initialize the main function if needed."""

    if __name__ == "__main__":
        cli()  # pylint: disable=no-value-for-parameter


maybe_init()



"""Base package for gate API code."""
import abc
import base64
import json
import secrets
import uuid
from datetime import datetime, timedelta
from hashlib import sha1
from typing import Dict, Generic, Optional, TypeVar, Union, cast
from xml.etree.ElementTree import Element  # nosec

from Crypto.Cipher import AES  # nosec
from Crypto.Cipher._mode_cbc import CbcMode  # nosec
from defusedxml import ElementTree
from httpx import AsyncClient, Response, RemoteProtocolError
from typing_extensions import Final

from .common import (
    CLOSE_DOOR_STATUSES,
    OPEN_DOOR_STATUSES,
    AllDoorStatus,
    ApiError,
    CachedTransitionDoorStatus,
    CredentialsIncorrectException,
    CredentialsNotSetException,
    DoorNotSetException,
    DoorStatus,
    ExceptionGenerator,
    GogoGate2ActivateResponse,
    GogoGate2InfoResponse,
    InvalidApiCodeException,
    InvalidDoorException,
    InvalidOptionException,
    ISmartGateActivateResponse,
    ISmartGateDoor,
    ISmartGateInfoResponse,
    RequestOption,
    TransitionDoorStatus,
    element_to_api_error,
    element_to_gogogate2_activate_response,
    element_to_gogogate2_info_response,
    element_to_ismartgate_activate_response,
    element_to_ismartgate_info_response,
    get_configured_doors,
    get_door_by_id,
)
from .const import GogoGate2ApiErrorCode, ISmartGateApiErrorCode

__author__ = "Robbie Van Gorkom; J. Nick Koston"
__email__ = "nick@koston.org"
# Do not edit this string manually, always use bumpversion
# Details in CONTRIBUTING.md
__version__ = "4.0.3"


def get_module_version():
    return __version__


class ApiCipher:
    """AES/CBC/PKCS5Padding algorithm."""

    def __init__(self, key: str) -> None:
        """Initialize the object."""
        self._key: Final = key
        self._key_bytes: Final = key.encode("utf-8")

    def encrypt(self, content: str, init_vector: Optional[str] = None) -> str:
        """Encrypt content."""
        init_vector_bytes: Final = ApiCipher.pad_pkcs5(
            init_vector or uuid.uuid4().hex
        ).encode("utf-8")[: AES.block_size]
        content_bytes: Final = ISmartGateApiCipher.pad_pkcs5(content).encode("utf-8")
        cipher: Final = cast(
            CbcMode, AES.new(self._key_bytes, AES.MODE_CBC, init_vector_bytes)
        )
        encrypted_bytes: Final = cipher.encrypt(content_bytes)
        return str(init_vector_bytes + base64.b64encode(encrypted_bytes), "utf-8")

    def decrypt(self, content: str) -> str:
        """Decrypt content."""
        init_vector: Final = content.encode("utf-8")[: AES.block_size]
        encrypted_bytes: Final = base64.b64decode(content[AES.block_size :])
        cipher: Final = cast(
            CbcMode, AES.new(self._key_bytes, AES.MODE_CBC, init_vector)
        )
        return ApiCipher.unpad_pkcs5(cipher.decrypt(encrypted_bytes)).decode("utf-8")

    @staticmethod
    def pad_pkcs5(data: str) -> str:
        """Add padding to bytes."""
        block_size: Final = AES.block_size
        return data + (block_size - len(data) % block_size) * chr(
            block_size - len(data) % block_size
        )

    @staticmethod
    def unpad_pkcs5(data: bytes) -> bytes:
        """Remove padding from bytes."""
        return data[0 : -data[-1]]


class GogoGate2ApiCipher(ApiCipher):
    """Cipher for GogoGate2 devices."""

    SHARED_SECRET: Final = "0e3b7%i1X9@54cAf"

    def __init__(self) -> None:
        """Initialize the object."""
        super().__init__(GogoGate2ApiCipher.SHARED_SECRET)


class ISmartGateApiCipher(ApiCipher):
    """Cipher for iSmartGate devices."""

    RAW_TOKEN_FORMAT: Final = "%s@ismartgate"

    def __init__(self, username: str, password: str) -> None:
        """Initialize the object."""
        self._username: Final = username
        self._password: Final = password

        # Calculate the token.
        raw_token: Final[str] = (
            ISmartGateApiCipher.RAW_TOKEN_FORMAT % self._username.lower()
        )
        self._token: Final = sha1(raw_token.encode("utf-8")).hexdigest()  # nosec

        # Calculate the key and pass it onto the superclass.
        sha1_hex_str: Final = sha1(  # nosec
            (self._username.lower() + self._password).encode()
        ).hexdigest()

        super().__init__(
            f"{sha1_hex_str[32:36]}a{sha1_hex_str[7:10]}!{sha1_hex_str[18:21]}*#{sha1_hex_str[24:26]}"
        )

    @property
    def token(self) -> str:
        """Get the token."""
        return self._token


ApiCipherType = TypeVar(
    "ApiCipherType", bound=Union[GogoGate2ApiCipher, ISmartGateApiCipher]
)
InfoResponseType = TypeVar(
    "InfoResponseType", bound=Union[GogoGate2InfoResponse, ISmartGateInfoResponse]
)
ActivateResponseType = TypeVar(
    "ActivateResponseType",
    bound=Union[GogoGate2ActivateResponse, ISmartGateActivateResponse],
)


# pylint: disable=too-many-instance-attributes
class AbstractGateApi(
    Generic[ApiCipherType, InfoResponseType, ActivateResponseType], abc.ABC
):
    """API capable of communicating with a gogogate2 devices."""

    API_URL_TEMPLATE: Final = "http://%s/api.php"
    DEFAULT_REQUEST_TIMEOUT = timedelta(seconds=20)
    DEFAULT_TRANSITION_STATUS_TIMEOUT = timedelta(seconds=55)

    def __init__(
        self,
        host: str,
        username: str,
        password: str,
        api_cipher: ApiCipherType,
        request_timeout: timedelta = DEFAULT_REQUEST_TIMEOUT,
        transition_status_timeout: timedelta = DEFAULT_TRANSITION_STATUS_TIMEOUT,
        httpx_async_client: Optional[AsyncClient] = None,
    ) -> None:
        """Initialize the object."""
        self._host: Final = host
        self._username: Final = username
        self._password: Final = password
        self._cipher: Final = api_cipher
        self._request_timeout: Final = request_timeout
        self._transition_status_timeout: Final = transition_status_timeout
        self._api_url: Final = AbstractGateApi.API_URL_TEMPLATE % host
        self._transition_door_status: Final[Dict[int, CachedTransitionDoorStatus]] = {}
        self._httpx_async_client: Final = httpx_async_client

    @property
    def host(self) -> str:
        """Get the host."""
        return self._host

    @property
    def username(self) -> str:
        """Get the username."""
        return self._username

    @property
    def password(self) -> str:
        """Get the password."""
        return self._password

    @property
    def cipher(self) -> ApiCipherType:
        """Get the cipher."""
        return self._cipher

    async def _async_request(
        self,
        option: RequestOption,
        arg1: Optional[str] = None,
        arg2: Optional[str] = None,
    ) -> Element:
        command_str: Final = json.dumps(
            (
                self._username,
                self._password,
                option.value,
                "" if arg1 is None else arg1,
                "" if arg2 is None else arg2,
            )
        )

        response: Final = await self._async_get_with_retry_on_protocol_error(
            params={
                "data": self._cipher.encrypt(command_str),
                **self._get_extra_url_params(),
            },
        )
        response_raw: Final = response.content.decode("utf-8")
        try:
            # Error messages are returned unencrypted so we try to decrypt the response.
            # if that fails, then we use what was returned.
            response_text = self._cipher.decrypt(response_raw)
        except ValueError:
            response_text = response_raw

        root_element: Final = ElementTree.fromstring(response_text)

        error_element: Final = root_element.find("error")
        if error_element:
            api_error: Final = element_to_api_error(error_element)
            raise self._get_exception_map().get(api_error.code, ApiError)(
                api_error.code, api_error.message
            )

        return cast(Element, root_element)

    async def _async_get_with_retry_on_protocol_error(
        self, params: Dict[str, str]
    ) -> Response:
        """Request the api url, and retry if the server disconnects on the first try."""
        client = self._httpx_async_client or AsyncClient()
        timeout = self._request_timeout.seconds
        url = self._api_url
        try:
            return await client.get(url, params=params, timeout=timeout)
        except RemoteProtocolError:
            # If the server disconnected without sending a response, we try again
            return await client.get(url, params=params, timeout=timeout)
        finally:
            if not self._httpx_async_client:
                await client.aclose()

    @abc.abstractmethod
    async def async_info(self) -> InfoResponseType:
        """Get info about the device and doors."""

    @abc.abstractmethod
    async def async_activate(self, door_id: int) -> ActivateResponseType:
        """Send a command to open/close/stop the door.

        Devices do not have a status for opening or closing. So running
        this method during an action will stop the door. It's recommended you
        use open_door() or close_door() as those methods check the status
        before running and run if needed."""

    @abc.abstractmethod
    def _get_activate_api_code(self, info: InfoResponseType, door_id: int) -> str:
        """Get api code for activate actions."""

    @staticmethod
    @abc.abstractmethod
    def _get_exception_map() -> Dict[int, ExceptionGenerator]:
        """Return a more specific exception."""

    # pylint: disable=no-self-use
    def _get_extra_url_params(self) -> Dict[str, str]:
        return {}

    async def _async_info(self) -> Element:
        """Get info about the device and doors."""
        return await self._async_request(RequestOption.INFO)

    async def _async_activate(
        self, door_id: int, info: Optional[InfoResponseType] = None
    ) -> Element:
        """Send a command to open/close/stop the door.

        Gogogate2/iSmartGate do not have a status for opening or closing. So
        running this method during an action will stop the door. It's
        recommended you use open_door() or close_door() as those methods check
        the status before running and run if needed."""
        return await self._async_request(
            RequestOption.ACTIVATE,
            str(door_id),
            self._get_activate_api_code(
                info if info else await self.async_info(), door_id
            ),
        )

    async def _async_set_door_status(
        self,
        door_id: int,
        target_door_status: DoorStatus,
        consider_transitional_states: bool = True,
    ) -> bool:
        """Send call to open/close a door if door is not already in that state."""
        if target_door_status == DoorStatus.UNDEFINED:
            return False

        # Get current door status.
        info: Final = await self.async_info()
        statuses: Final = self._get_door_statuses(
            info, use_transitional_status=consider_transitional_states
        )
        current_door_status: Final = statuses.get(door_id)
        result_door_statuses: Final = (
            OPEN_DOOR_STATUSES
            if target_door_status == DoorStatus.OPENED
            else CLOSE_DOOR_STATUSES
        )
        transitional_door_status: Final = (
            TransitionDoorStatus.OPENING
            if target_door_status == DoorStatus.OPENED
            else TransitionDoorStatus.CLOSING
        )

        # Door is invalid, not configured, already in desired state or transitioning to it.
        if not current_door_status or current_door_status in result_door_statuses:
            return False

        await self._async_activate(door_id, info)

        # Set the transitional status.
        self._transition_door_status[door_id] = CachedTransitionDoorStatus(
            door_id=door_id,
            activated=datetime.utcnow(),
            transition_status=transitional_door_status,
            target_status=target_door_status,
        )

        return True

    async def async_close_door(
        self, door_id: int, consider_transitional_states: bool = True
    ) -> bool:
        """Close a door.

        :return True if close command sent, False otherwise.
        """
        return await self._async_set_door_status(
            door_id,
            DoorStatus.CLOSED,
            consider_transitional_states=consider_transitional_states,
        )

    async def async_open_door(
        self, door_id: int, consider_transitional_states: bool = True
    ) -> bool:
        """Open a door.

        :return True if open command sent, False otherwise.
        """
        return await self._async_set_door_status(
            door_id,
            DoorStatus.OPENED,
            consider_transitional_states=consider_transitional_states,
        )

    def _get_door_statuses(
        self, info: InfoResponseType, use_transitional_status: bool = True
    ) -> Dict[int, AllDoorStatus]:
        doors: Final = get_configured_doors(info)

        # Clean out the cache.
        for (
            cached_door_id,
            cached_transitional_status,
        ) in list(self._transition_door_status.items()):
            if (
                datetime.utcnow() - cached_transitional_status.activated
                >= self._transition_status_timeout
            ):
                del self._transition_door_status[cached_door_id]

        # For each door, determine the status.
        result: Final[Dict[int, AllDoorStatus]] = {}
        for door in doors:
            transitional_status = self._transition_door_status.get(door.door_id)
            if (
                use_transitional_status
                and transitional_status
                and transitional_status.target_status != door.status
            ):
                result[door.door_id] = transitional_status.transition_status
            else:
                result[door.door_id] = door.status

        return result

    async def async_get_door_statuses(
        self, use_transitional_status: bool = True
    ) -> Dict[int, AllDoorStatus]:
        """Get configured door statuses."""
        return self._get_door_statuses(
            await self.async_info(), use_transitional_status=use_transitional_status
        )

    def async_get_door_statuses_from_info(
        self, info: InfoResponseType, use_transitional_status: bool = True
    ) -> Dict[int, AllDoorStatus]:
        """Get configured door statuses from an existing info response."""
        return self._get_door_statuses(
            info, use_transitional_status=use_transitional_status
        )


class ISmartGateApi(
    AbstractGateApi[
        ISmartGateApiCipher, ISmartGateInfoResponse, ISmartGateActivateResponse
    ]
):
    """API for interacting with iSmartGate devices."""

    def __init__(
        self,
        host: str,
        username: str,
        password: str,
        request_timeout: timedelta = AbstractGateApi.DEFAULT_REQUEST_TIMEOUT,
        transition_status_timeout: timedelta = AbstractGateApi.DEFAULT_TRANSITION_STATUS_TIMEOUT,
        httpx_async_client: Optional[AsyncClient] = None,
    ) -> None:
        """Initialize the object."""
        super().__init__(
            host,
            username,
            password,
            ISmartGateApiCipher(username, password),
            request_timeout=request_timeout,
            transition_status_timeout=transition_status_timeout,
            httpx_async_client=httpx_async_client,
        )

    async def async_info(self) -> ISmartGateInfoResponse:
        """Get info about the device and doors."""
        return element_to_ismartgate_info_response(await self._async_info())

    async def async_activate(self, door_id: int) -> ISmartGateActivateResponse:
        """Send a command to open/close/stop the door.

        Devices do not have a status for opening or closing. So running
        this method during an action will stop the door. It's recommended you
        use open_door() or close_door() as those methods check the status
        before running and run if needed."""
        return element_to_ismartgate_activate_response(
            await self._async_activate(door_id)
        )

    #  pylint: disable=no-self-use
    def _get_activate_api_code(self, info: ISmartGateInfoResponse, door_id: int) -> str:
        """Get api code for activate actions."""
        door: Final = get_door_by_id(door_id, info)
        if not door:
            raise InvalidDoorException(door_id)

        return cast(ISmartGateDoor, door).apicode

    @staticmethod
    def _get_exception_map() -> Dict[int, ExceptionGenerator]:
        """Return a more specific exception."""
        return {
            ISmartGateApiErrorCode.CREDENTIALS_NOT_SET.value: CredentialsNotSetException,
            ISmartGateApiErrorCode.CREDENTIALS_INCORRECT.value: CredentialsIncorrectException,
            ISmartGateApiErrorCode.INVALID_OPTION.value: InvalidOptionException,
            ISmartGateApiErrorCode.INVALID_API_CODE.value: InvalidApiCodeException,
            ISmartGateApiErrorCode.DOOR_NOT_SET.value: DoorNotSetException,
        }

    def _get_extra_url_params(self) -> Dict[str, str]:
        """Get extra url params when making calls."""
        return {
            "t": str(secrets.randbelow(100000000) + 1),
            "token": self.cipher.token,
        }


class GogoGate2Api(
    AbstractGateApi[
        GogoGate2ApiCipher, GogoGate2InfoResponse, GogoGate2ActivateResponse
    ]
):
    """API for interacting with GogoGate2 devices."""

    def __init__(
        self,
        host: str,
        username: str,
        password: str,
        request_timeout: timedelta = AbstractGateApi.DEFAULT_REQUEST_TIMEOUT,
        transition_status_timeout: timedelta = AbstractGateApi.DEFAULT_TRANSITION_STATUS_TIMEOUT,
        httpx_async_client: Optional[AsyncClient] = None,
    ) -> None:
        """Initialize the object."""
        super().__init__(
            host,
            username,
            password,
            GogoGate2ApiCipher(),
            request_timeout=request_timeout,
            transition_status_timeout=transition_status_timeout,
            httpx_async_client=httpx_async_client,
        )

    async def async_info(self) -> GogoGate2InfoResponse:
        """Get info about the device and doors."""
        return element_to_gogogate2_info_response(await self._async_info())

    async def async_activate(self, door_id: int) -> GogoGate2ActivateResponse:
        """Send a command to open/close/stop the door.

        Devices do not have a status for opening or closing. So running
        this method during an action will stop the door. It's recommended you
        use open_door() or close_door() as those methods check the status
        before running and run if needed."""
        return element_to_gogogate2_activate_response(
            await self._async_activate(door_id)
        )

    @staticmethod
    def _get_exception_map() -> Dict[int, ExceptionGenerator]:
        """Return a more specific exception."""
        return {
            GogoGate2ApiErrorCode.CREDENTIALS_NOT_SET.value: CredentialsNotSetException,
            GogoGate2ApiErrorCode.CREDENTIALS_INCORRECT.value: CredentialsIncorrectException,
            GogoGate2ApiErrorCode.INVALID_OPTION.value: InvalidOptionException,
            GogoGate2ApiErrorCode.INVALID_API_CODE.value: InvalidApiCodeException,
            GogoGate2ApiErrorCode.DOOR_NOT_SET.value: DoorNotSetException,
        }

    # pylint: disable=unused-argument, no-self-use
    def _get_activate_api_code(self, info: GogoGate2InfoResponse, door_id: int) -> str:
        """Get api code for activate actions."""
        return info.apicode



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import find_packages, setup

with open("README.md") as readme_file:
    readme = readme_file.read()

setup_requirements = [
    "pytest-runner>=5.2",
]

test_requirements = [
    "typing_extensions>=3.10.0.0",
    "asynctest>=0.13.0",
    "respx>=0.17.0",
    "pytest-asyncio",
    "black>=19.10b0",
    "codecov>=2.1.4",
    "flake8>=3.8.3",
    "flake8-debugger>=3.2.1",
    "pytest>=5.4.3",
    "pytest-cov>=2.9.0",
    "pytest-raises>=0.11",
]

dev_requirements = [
    *setup_requirements,
    *test_requirements,
    "bump2version>=1.0.1",
    "coverage>=5.1",
    "ipython>=7.15.0",
    "m2r2>=0.2.7",
    "pytest-runner>=5.2",
    "Sphinx>=3.4.3",
    "sphinx_rtd_theme>=0.5.1",
    "tox>=3.15.2",
    "twine>=3.1.1",
    "wheel>=0.34.2",
]

requirements = [
    "dicttoxml>=1.7.4",
    "defusedxml>=0.7.1",
    "pycryptodome>=3.9.8",
    "httpx>=0.18.0",
]

extra_requirements = {
    "setup": setup_requirements,
    "test": test_requirements,
    "dev": dev_requirements,
    "all": [
        *requirements,
        *dev_requirements,
    ],
}

setup(
    author="Robbie Van Gorkom; J. Nick Koston",
    author_email="nick@koston.org",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    description="iSmartGate and GogoGate2 API",
    entry_points={
        "console_scripts": [
            "ismartgate=ismartgate.cli:ismartgate_cli",
            "gogogate2=ismartgate.cli:gogogate2_cli",
        ],
    },
    install_requires=requirements,
    license="MIT license",
    long_description=readme,
    long_description_content_type="text/markdown",
    include_package_data=True,
    keywords="ismartgate",
    name="ismartgate",
    packages=find_packages(exclude=["tests", "*.tests", "*.tests.*"]),
    python_requires=">=3.7",
    setup_requires=setup_requirements,
    test_suite="ismartgate/tests",
    tests_require=test_requirements,
    extras_require=extra_requirements,
    url="https://github.com/bdraco/ismartgate",
    # Do not edit this string manually, always use bumpversion
    # Details in CONTRIBUTING.rst
    version="4.0.3",
    zip_safe=False,
)
